/* 	$NetBSD$ */

/*-
 * Copyright (c) 2007 Jachym Holecek <freza@NetBSD.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/resource.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <prop/proplib.h>

static const char *path;
static const char *name;
static u_long nloops = 10000;

static double time_ioctl_xml = 0.0;
static double time_ioctl_bin = 0.0;
static double time_intern_xml = 0.0;
static double time_intern_bin = 0.0;
static double time_extern_xml = 0.0;
static double time_extern_bin = 0.0;
static size_t size_extern_xml;
static size_t size_extern_bin;

/*
 * Utilities.
 */
static prop_object_t
internalize(const char *path)
{
	prop_object_t po;

	po = prop_dictionary_internalize_from_file(path);
	if (po == NULL)
		po = prop_array_internalize_from_file(path);

	return (po);
}

static prop_object_t
parsexml(uint8_t *buf)
{
	const char *s = (const char *)buf;
	prop_object_t po;

	po = prop_dictionary_internalize(s);
	if (po == NULL)
		po = prop_array_internalize(s);

	return (po);
}

static prop_object_t
parsebin(uint8_t *buf, size_t len)
{
	prop_object_t po;

	po = prop_dictionary_binary_internalize(buf, len);
	if (po == NULL)
		po = prop_array_binary_internalize(buf, len);

	return (po);
}

static bool
blistify(prop_object_t po, uint8_t **buf, size_t *len)
{
	if (!prop_dictionary_binary_externalize(po, buf, len) &&
	    !prop_array_binary_externalize(po, buf, len))
		return (false);

	return (true);
}

static uint8_t *
xmlify(prop_object_t po)
{
	char *s;

	s = prop_dictionary_externalize(po);
	if (s == NULL)
		s = prop_array_externalize(po);

	return ((uint8_t *)s);
}

double
getpoint(void)
{
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru) == -1)
		err(1, "getrusage");

	return ((double)ru.ru_utime.tv_sec +
		(double)ru.ru_utime.tv_usec/1000000.0);
}

static void
failed(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "FAILED %s ", name);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	fprintf(stderr, "\n");
	exit(EXIT_FAILURE);
}

/*
 * Regression test, assumes input file is XML.
 */

static void
regression(void)
{
	prop_object_t po1; /* plist from XML file */
	prop_object_t po2;
	uint8_t *buf;
	size_t len;

	po1 = internalize(path);
	if (po1 == NULL)
		failed("internalize XML file");

	buf = xmlify(po1);
	if (buf == NULL)
		failed("externalize XML");

	po2 = parsexml(buf);
	if (po2 == NULL)
		failed("internalize XML");
	free(buf);

	if (! prop_object_equals(po1, po2))
		failed("compare XML");
	prop_object_release(po2);

	if (! blistify(po1, &buf, &len))
		failed("externalize binary");

	po2 = parsebin(buf, len);
	if (po2 == NULL)
		failed("internalize binary");

	if (! prop_object_equals(po1, po2))
		failed("compare binary");
	prop_object_release(po2);

	prop_object_release(po1);
}

static void
benchmark(void)
{
	prop_object_t po1;
	prop_object_t po2;
	uint8_t *buf;
	uint8_t *tmp;
	size_t size;
	double beg;
	int i;

	po1 = internalize(path);
	if (po1 == NULL)
		failed("benchmark: internalize XML file");

	/* How long does XML internalize take? */
	buf = xmlify(po1);
	if (buf == NULL)
		failed("benchmark: prepare XML");
	size_extern_xml = strlen((char *)buf);

	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		po2 = parsexml(buf);
		if (po2 == NULL)
			failed("benchmark: internalize XML");

		time_intern_xml += (getpoint() - beg);
		prop_object_release(po2);
	}
	free(buf);

	/* How long does binary internalize take? */
	if (! blistify(po1, &buf, &size))
		failed("benchmark: prepare binary");
	size_extern_bin = size;

	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		po2 = parsebin(buf, size);
		if (po2 == NULL)
			failed("benchmark: internalize binary");

		time_intern_bin += (getpoint() - beg);
		prop_object_release(po2);
	}
	free(buf);

	/* How long does XML externalize take? */
	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		buf = xmlify(po1);
		if (buf == NULL)
			failed("benchmark: externalize XML");

		time_extern_xml += (getpoint() - beg);
		free(buf);
	}

	/* How long does binary externalize take? */
	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		if (! blistify(po1, &buf, &size))
			failed("benchmark: externalize binary");

		time_extern_bin += (getpoint() - beg);
		free(buf);
	}

	/* How long does full XML transport take? */
	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		buf = xmlify(po1);
		if (buf == NULL)
			failed("benchmark: transport XML send");

		size = (1 + strlen((const char *)buf));
		tmp = malloc(size);
		if (tmp == NULL)
			failed("benchmark: transport XML buffer alloc");

		memcpy(tmp, buf, size);
		po2 = parsexml(tmp);
		if (po2 == NULL)
			failed("benchmark: transport XML recv");

		time_ioctl_xml += (getpoint() - beg);
		prop_object_release(po2);
		free(buf);
		free(tmp);
	}

	/* How long does full binary transport take? */
	for (i = 0; i < nloops; i++) {
		beg = getpoint();

		if (! blistify(po1, &buf, &size))
			failed("benchmark: transport binary send");

		tmp = malloc(size);
		if (tmp == NULL)
			failed("benchmark: transport binary buffer alloc");

		memcpy(tmp, buf, size);
		po2 = parsebin(tmp, size);
		if (po2 == NULL)
			failed("benchmark: transport binary recv");

		time_ioctl_bin += (getpoint() - beg);
		prop_object_release(po2);
		free(buf);
		free(tmp);
	}
}

static double
percentage(double big, double small)
{
	double one = (big / 100);

	return (small / one);
}

int
xstrtoul(char *str, u_long *val)
{
	char                    *end;
	int 			base = 10;
	u_long 			v;

	if (str[0] == '0')
		switch (str[1]) {
		case 'x':
			base = 16;
			str += 2;
			break;
		case 'd':
			base = 10;
			str += 2;
			break;
		case 'o':
			base = 8;
			str += 2;
			break;
		case 'b':
			base = 2;
			str += 2;
			break;
		}

	v = (u_long) strtoul(str, &end, base);
	if (*end != '\0' || str[0] == '\0')
		return (EINVAL);

	*val = v;
	return (0);
}

int
main(int argc, char *argv[])
{
	double val1;
	double val2;
	int verbose = 0;
	int c;

	while ((c = getopt(argc, argv, "n:v")) != -1)
		switch (c) {
		case 'n':
			if (xstrtoul(optarg, &nloops))
				errx(EXIT_FAILURE, "integer expected");
			break;

		case 'v':
			verbose = 1;
			break;

		default:
			errx(EXIT_FAILURE, "invalid or misused option -%c",
			    optopt);
		}

	argc -= optind;
	argv += optind;

	if (argc != 1)
		errx(1, "need exactly one XML plist file argument");

	path = argv[0];
	name = basename(argv[0]);

	regression();
	benchmark();

	printf("PASSED %s ", name);


	if (verbose)
		printf("(XML %.2fs %.2fs %.2f %uB BIN %.2fs %.2fs %.2f %uB)",
		    time_intern_xml, time_extern_xml, time_ioctl_xml,
		    (u_int)size_extern_xml,
		    time_intern_bin, time_extern_bin, time_ioctl_bin,
		    (u_int)size_extern_bin);
	else {
		/* Relative results for full transport benchmark. */
		if (time_ioctl_xml == time_ioctl_bin)
			printf("TRANSPORT even ");
		else
		if (time_ioctl_xml > time_ioctl_bin)
			printf("TRANSPORT bin %.0f%% ",
			    percentage(time_ioctl_xml, time_ioctl_bin));
		else
			printf("TRANSPORT xml %.0f%% ",
			    percentage(time_ioctl_bin, time_ioctl_xml));

		/* Relative results for raw intern/extern benchmark. */
		val1 = (time_intern_xml + time_extern_xml);
		val2 = (time_intern_bin + time_extern_bin);
		if (val1 == val2)
			printf("RAW even ");
		else
		if (val1 > val2)
			printf("RAW bin %.0f%% ", percentage(val1, val2));
		else
			printf("RAW xml %.0f%% ", percentage(val2, val1));
	
		/* Relative results for encoding size benchmark. */
		if (size_extern_bin == size_extern_xml)
			printf("SIZE even ");
		else
		if (size_extern_xml > size_extern_bin)
			printf("SIZE %.0f%% ",
			    percentage(size_extern_xml, size_extern_bin));
		else
			printf("SIZE %.0f%% ",
			    percentage(size_extern_bin, size_extern_xml));
	}

	printf("\n");
	return (0);
}
