/* 	$NetBSD$ */

/*-
 * Copyright (c) 2007 Jachym Holecek <freza@NetBSD.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <prop/proplib.h>

static prop_object_t
internalize(const char *path)
{
	prop_object_t po;

	po = prop_dictionary_internalize_from_file(path);
	if (po == NULL) {
		po = prop_array_internalize_from_file(path);
		if (po == NULL)
			errx(EXIT_FAILURE, "could not internalize %s", path);
	}

	return (po);
}

static void
externalize(int fd, prop_object_t po, const char *format)
{
	uint8_t *buf;
	size_t len;
	char *s;

	if (strcasecmp(format, "xml") == 0) {
		s = prop_dictionary_externalize(po);
		if (s == NULL) {
			s = prop_array_externalize(po);
			if (s == NULL)
				errx(EXIT_FAILURE, "could not externalize");
		}

		buf = (uint8_t *)s;
		len = strlen(s);
	} else
	if (strcasecmp(format, "bplist") == 0) {
		if (!prop_dictionary_binary_externalize(po, &buf, &len) &&
		    !prop_array_binary_externalize(po, &buf, &len))
			errx(EXIT_FAILURE, "could not externalize");
	}

	/* Not prop_*_externalize_to_file(), stdout may be pipe/terminal. */
	if (write(fd, buf, len) == -1)
		err(EXIT_FAILURE, "could not write");
}

#define OPTIONS 		"ho:O:"

static void
usage(void)
{
	const char *me = getprogname();

	printf("Usage: %s [-o file] [-O format] file\n", me);
	printf("       %s file0 file1\n", me);
	printf("-h        \t Print (this) help\n");
	printf("-o file   \t Output file [stdout]\n");
	printf("-O format \t Externalize format [xml bplist]\n");

	exit(EXIT_SUCCESS);
}

int
main(int argc, char *argv[])
{
	const char *oformat = "xml";
	const char *opath = NULL;
	int fdo = STDOUT_FILENO;
	int c;

	while ((c = getopt(argc, argv, OPTIONS)) != -1)
		switch (c) {
		case 'h': /* Help. */
			usage();
			/* UNREACHED */

		case 'o': /* Output file. */
			opath = optarg;
			break;

		case 'O': /* Externalize format. */
			oformat = optarg;
			break;

		case '?':
			errx(EXIT_FAILURE, "uknown option -%c", optopt);
			/* UNREACHED */

		case ':':
			errx(EXIT_FAILURE, "missing argument to -%c", optopt);
			/* UNREACHED */
		}

	argc -= optind;
	argv += optind;

	if (argc == 2) {
		if (prop_object_equals(internalize(argv[0]),
		    internalize(argv[1])))
			return (EXIT_SUCCESS);

		return (EXIT_FAILURE);
	} else
	if (argc == 1) {
		prop_object_t po;

		po = internalize(argv[0]);

		if (opath != NULL) {
			fdo = open(opath, O_WRONLY|O_CREAT|O_TRUNC, 0644);
			if (fdo == -1)
				errx(EXIT_FAILURE, "could not open %s", opath);
		}

		externalize(fdo, po, oformat);
		return (EXIT_SUCCESS);
	}

	warnx("unknown mode of operation");
	return (EXIT_FAILURE);
}
